package chapter9.apigateway.client.gui;

import java.util.List;
import java.util.Optional;
import chapter9.apigateway.client.model.Cd;
import chapter9.apigateway.client.remote.RemoteRequest;
import chapter9.apigateway.client.remote.RemoteRequestException;
import javafx.application.Platform;
import javafx.collections.*;
import javafx.geometry.Insets;
import javafx.scene.layout.*;
import javafx.scene.control.*;

public class View
{
    private BorderPane pane;
    private ListView<Cd> list;
    private AddChangeDialog addChangeDialog;
    private Alert noSelectionAlert, confirmationAlert, remoteRequestAlert;

    public View(ObservableList<Cd> data)
    {
        list = new ListView<>(data);

        HBox controls = new HBox(10);
        controls.setPadding(new Insets(10, 10, 10, 10));
        Button add = new Button("Hinzufgen");
        add.setOnAction(e->showAddDialog());
        Button change = new Button("ndern");
        change.setOnAction(e->showChangeDialog());
        Button delete = new Button("Lschen");
        delete.setOnAction(e->delete());
        Button reload = new Button("Neu laden");
        reload.setOnAction(e->reload());
        Button sortArtist = new Button("Sortieren nach Knstler");
        sortArtist.setOnAction(e->sortArtist());
        Button sortTitle = new Button("Sortieren nach Titel");
        sortTitle.setOnAction(e->sortTitle());
        controls.getChildren().addAll(add, change, delete, reload, sortArtist, sortTitle);

        pane = new BorderPane();
        pane.setPadding(new Insets(10, 10, 10, 10));
        pane.setCenter(list);
        pane.setBottom(controls);

        addChangeDialog = new AddChangeDialog();

        noSelectionAlert = new Alert(Alert.AlertType.ERROR);
        noSelectionAlert.setTitle("Fehler");
        noSelectionAlert.setHeaderText(null);
        noSelectionAlert.setContentText("Es ist kein Eintrag in der Tabelle ausgewhlt.");

        confirmationAlert = new Alert(Alert.AlertType.CONFIRMATION);
        confirmationAlert.setTitle("Besttigung");
        confirmationAlert.setHeaderText(null);
        confirmationAlert.setContentText("Wollen Sie den Eintrag wirklich lschen?");

        remoteRequestAlert = new Alert(Alert.AlertType.ERROR);
        remoteRequestAlert.setTitle("Fehler");
        remoteRequestAlert.setHeaderText(null);
    }

    private void showAddDialog()
    {
        addChangeDialog.prepareAddOperation();
        Cd cd = addChangeDialog.showDialog();
        if(cd != null)
        {
            new Thread(()->asynchPost(cd)).start();
        }
    }
    
    private void asynchPost(Cd cd)
    {
        try
        {
            String response = RemoteRequest.post(cd.toJson());
            if(response.startsWith("{id:") && response.endsWith("}"))
            {
                String id = response.substring(4);
                id = id.substring(0, id.length()-1);
                id = id.trim();
                cd.setId(id);
            }
            Platform.runLater(()->list.getItems().add(cd));
        }
        catch(RemoteRequestException e)
        {
            Platform.runLater(()->
            {
                remoteRequestAlert.setContentText(e.toString());
                remoteRequestAlert.showAndWait();
            });
        }
    }

    private void showChangeDialog()
    {
        Cd oldCd = list.getSelectionModel().getSelectedItem();
        if(oldCd != null)
        {
            addChangeDialog.prepareChangeOperation(oldCd);
            Cd newCd = addChangeDialog.showDialog();
            if(newCd != null)
            {
                new Thread(()->asynchPut(oldCd, newCd)).start();
            }
        }
        else
        {
            noSelectionAlert.showAndWait();
        }
    }
    
    private void asynchPut(Cd oldCd, Cd newCd)
    {
        try
        {
            RemoteRequest.put(oldCd.getId(), newCd.toJson());
            Platform.runLater(()->
            {
                newCd.setId(oldCd.getId());
                ObservableList<Cd> data = list.getItems();
                int index = data.indexOf(oldCd);
                data.set(index, newCd);
            });
        }
        catch(RemoteRequestException e)
        {
            Platform.runLater(()->
            {
                remoteRequestAlert.setContentText(e.toString());
                remoteRequestAlert.showAndWait();
            });
        }
    }

    private void delete()
    {
        Cd cd = list.getSelectionModel().getSelectedItem();
        if(cd != null)
        {
            Optional<ButtonType> result = confirmationAlert.showAndWait();
            if(result.isPresent())
            {
                ButtonType bt = result.get();
                if(bt == ButtonType.OK)
                {
                    new Thread(()->asynchDelete(cd)).start();
                }
            }
        }
        else
        {
            noSelectionAlert.showAndWait();
        }
    }
    
    private void asynchDelete(Cd cd)
    {
        try
        {
            RemoteRequest.delete(cd.getId());
            Platform.runLater(()->list.getItems().remove(cd));
        }
        catch(RemoteRequestException e)
        {
            Platform.runLater(()->
            {
                remoteRequestAlert.setContentText(e.toString());
                remoteRequestAlert.showAndWait();
            });
        }
    }

    private void reload()
    {
        new Thread(()->asynchReload()).start();
    }
    
    private void asynchReload()
    {
        try
        {
            List<Cd> rawData = Main.fill();
            Platform.runLater(()->list.setItems(FXCollections.observableList(rawData)));
        }
        catch(RemoteRequestException e)
        {
            Platform.runLater(()->
            {
                remoteRequestAlert.setContentText(e.toString());
                remoteRequestAlert.showAndWait();
            });
        }
    }

    private void sortArtist()
    {
        ObservableList<Cd> basicData = list.getItems();
        basicData.sort((Cd cd1, Cd cd2) -> cd1.getArtist().compareTo(cd2.getArtist()));
    }

    private void sortTitle()
    {
        ObservableList<Cd> basicData = list.getItems();
        basicData.sort((Cd cd1, Cd cd2) -> cd1.getTitle().compareTo(cd2.getTitle()));
    }

    public Pane getUI()
    {
        return pane;
    }
}