package chapter9.apigateway.lambda;

import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import org.json.simple.*;
import org.json.simple.parser.*;
import com.amazonaws.services.dynamodbv2.*;
import com.amazonaws.services.dynamodbv2.model.*;
import com.amazonaws.services.lambda.runtime.*;
import com.google.gson.*;

public class LambdaFunctionHandler implements RequestStreamHandler
{
    private static final String TABLE_NAME = "Oechsle_CD_Table";
    
    @SuppressWarnings("unchecked")
    public void handleRequest(InputStream input, OutputStream output, Context context) throws IOException
    {
        BufferedReader reader = new BufferedReader(new InputStreamReader(input));
        JSONParser parser = new JSONParser();
        JSONObject responseJson = new JSONObject();

        try
        {
            JSONObject event = (JSONObject) parser.parse(reader);
            context.getLogger().log(event.toString());
            String method = (String) event.get("httpMethod");
            if(method.equals("GET"))
            {
                doGet(event, context, responseJson);
            }
            else if(method.equals("POST"))
            {
                doPost(event, context, responseJson);
            }
            else if(method.equals("PUT"))
            {
                doPut(event, context, responseJson);
            }
            else if(method.equals("DELETE"))
            {
                doDelete(event, context, responseJson);
            }
            else
            {
                responseJson.put("statusCode", 400);
                responseJson.put("cause", "bad request method: " + method);
            }

            JSONObject headerJson = new JSONObject();
            headerJson.put("x-custom-header", "my custom header value");

            responseJson.put("headers", headerJson);

        }
        catch(ParseException pex)
        {
            responseJson.put("statusCode", 400);
            responseJson.put("exception", pex);
        }

        OutputStreamWriter writer = new OutputStreamWriter(output, "UTF-8");
        writer.write(responseJson.toString());
        writer.close();
    }

    @SuppressWarnings("unchecked")
    private void doGet(JSONObject event, Context context, JSONObject responseJson)
    {
        String result = null;
        
        JSONObject responseBody = new JSONObject();

        if(event.get("pathParameters") != null)
        {
            context.getLogger().log("pathParameters found");
            JSONObject pps = (JSONObject) event.get("pathParameters");
            String id = (String)pps.get("id"); 
            if(id != null)
            {
                result = get(id);
            }
        }
        else if(event.get("queryStringParameters") != null)
        {
            context.getLogger().log("queryStringParameters found");
            JSONObject qps = (JSONObject) event.get("queryStringParameters");
            String id = (String)qps.get("id"); 
            if(id != null)
            {
                result = get(id);
            }
        }
        else
        {
            result = getAll();
        }
        if(result != null)
        {
            responseBody.put("result", result);
            responseJson.put("statusCode", 200);
            responseJson.put("body", responseBody.toString());
        }
        else
        {
            responseBody.put("message", "No item found");
            responseJson.put("statusCode", 404);
        }
    }
    
    private String get(String id)
    {
        AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().build();
        HashMap<String, AttributeValue> key = new HashMap<String, AttributeValue>();
        key.put("id", new AttributeValue(id));
        GetItemResult getResult = dynamoDB.getItem(TABLE_NAME, key);
        return getResult.getItem().toString();
    }
    
    private String getAll()
    {
        AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().build();
        ScanRequest req = new ScanRequest().withTableName(TABLE_NAME);
        ScanResult scanResult = dynamoDB.scan(req);
        return scanResult.toString();
    }

    @SuppressWarnings("unchecked")
    private void doPost(JSONObject event, Context context, JSONObject responseJson)
    {
        String result = null;
        
        JSONObject responseBody = new JSONObject();

        String data = (String)event.get("body");
        Map<String, String> bodyMap = new Gson().fromJson(data, new HashMap<String, Object>().getClass());

        AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().build();
        String id = System.currentTimeMillis() + "-" + System.nanoTime();
        HashMap<String, AttributeValue> attributes = new HashMap<String, AttributeValue>();
        attributes.put("id", new AttributeValue(id));
        for(Entry<String, String> entry: bodyMap.entrySet())
        {
            if(!entry.getKey().equals("id"))
            {
                attributes.put(entry.getKey(), new AttributeValue(entry.getValue()));
            }
        }
        PutItemResult putResult = dynamoDB.putItem(TABLE_NAME, attributes);
        result = putResult.toString();

        if(result != null)
        {
            responseBody.put("result", result);
            responseJson.put("statusCode", 200);
            responseJson.put("body", "{id: " + id + "}");
        }
        else
        {
            responseBody.put("message", "No item found");
            responseJson.put("statusCode", 404);
        }
    }

    @SuppressWarnings("unchecked")
    private void doPut(JSONObject event, Context context, JSONObject responseJson)
    {
        String result = null;
        
        JSONObject responseBody = new JSONObject();

        if(event.get("pathParameters") != null)
        {
            context.getLogger().log("pathParameters found");
            JSONObject pps = (JSONObject) event.get("pathParameters");
            String id = (String)pps.get("id"); 
            if(id != null)
            {
                String data = (String)event.get("body");
                Map<String, String> bodyMap = new Gson().fromJson(data, new HashMap<String, Object>().getClass());

                AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().build();
                HashMap<String, AttributeValue> attributes =
                                new HashMap<String, AttributeValue>();
                attributes.put("id", new AttributeValue(id));
                for(Entry<String, String> entry: bodyMap.entrySet())
                {
                    if(!entry.getKey().equals("id"))
                    {
                        attributes.put(entry.getKey(), new AttributeValue(entry.getValue()));
                    }
                }
                PutItemResult putResult = dynamoDB.putItem(TABLE_NAME, attributes);
                result = putResult.toString();
            }
        }
        if(result != null)
        {
            responseBody.put("result", result);
            responseJson.put("statusCode", 200);
            responseJson.put("body", responseBody.toString());
        }
        else
        {
            responseBody.put("message", "No item found");
            responseJson.put("statusCode", 404);
        }
    }

    @SuppressWarnings("unchecked")
    private void doDelete(JSONObject event, Context context, JSONObject responseJson)
    {
        String result = null;
        
        JSONObject responseBody = new JSONObject();

        if(event.get("pathParameters") != null)
        {
            context.getLogger().log("pathParameters found");
            JSONObject pps = (JSONObject) event.get("pathParameters");
            String id = (String)pps.get("id"); 
            if(id != null)
            {
                result = delete(id);
            }
        }
        else if(event.get("queryStringParameters") != null)
        {
            context.getLogger().log("queryStringParameters found");
            JSONObject qps = (JSONObject) event.get("queryStringParameters");
            String id = (String)qps.get("id"); 
            if(id != null)
            {
                result = delete(id);
            }
        }
        if(result != null)
        {
            responseBody.put("result", result);
            responseJson.put("statusCode", 200);
            responseJson.put("body", responseBody.toString());
        }
        else
        {
            responseBody.put("message", "No item found");
            responseJson.put("statusCode", 404);
        }
    }
    
    private String delete(String id)
    {
        AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().build();
        HashMap<String, AttributeValue> key =
                        new HashMap<String, AttributeValue>();
        key.put("id", new AttributeValue(id));
        DeleteItemResult deleteResult = dynamoDB.deleteItem(TABLE_NAME, key);
        return deleteResult.toString();
    }
}
