package chapter9.dynamodb;

/*
 * Copyright 2012-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  http://aws.amazon.com/apache2.0
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
import java.util.*;

import com.amazonaws.*;
import com.amazonaws.auth.profile.*;
import com.amazonaws.regions.Regions;
import com.amazonaws.services.dynamodbv2.*;
import com.amazonaws.services.dynamodbv2.document.*;
import com.amazonaws.services.dynamodbv2.model.*;
import com.amazonaws.services.dynamodbv2.util.TableUtils;

/**
 * This sample demonstrates how to perform a few simple operations with the
 * Amazon DynamoDB service.
 */
public class AmazonDynamoDBSample
{
    private static final String TABLE_NAME = "my-favorite-cd-table";

    /*
     * Before running the code: Fill in your AWS access credentials in the
     * provided credentials file template, and be sure to move the file to the
     * default location (C:\\Users\\oechsle\\.aws\\credentials) where the sample
     * code will load the credentials from.
     * https://console.aws.amazon.com/iam/home?#security_credential
     *
     * WARNING: To avoid accidental leakage of your credentials, DO NOT keep the
     * credentials file in your source directory.
     */

    /**
     * The only information needed to create a client are security credentials
     * consisting of the AWS Access Key ID and Secret Access Key. All other
     * configuration, such as the service endpoints, are performed
     * automatically. Client parameters, such as proxies, can be specified in an
     * optional ClientConfiguration object when constructing a client.
     *
     * @see com.amazonaws.auth.BasicAWSCredentials
     * @see com.amazonaws.auth.ProfilesConfigFile
     * @see com.amazonaws.ClientConfiguration
     */
    public static void main(String[] args) throws Exception
    {
        /*
         * The ProfileCredentialsProvider will return your [default] credential
         * profile by reading from the credentials file located at
         * (C:\\Users\\oechsle\\.aws\\credentials).
         */
        ProfileCredentialsProvider credentialsProvider = new ProfileCredentialsProvider("DynamoDB");
        try
        {
            credentialsProvider.getCredentials();
        }
        catch(Exception e)
        {
            throw new AmazonClientException("Login-Daten nicht gefunden.", e);
        }
        AmazonDynamoDB dynamoDB = AmazonDynamoDBClientBuilder.standard().withCredentials(credentialsProvider).withRegion(Regions.EU_CENTRAL_1).build();

        try
        {
            // Create a table with a primary hash key named 'name', which holds
            // a string
            CreateTableRequest createTableRequest = new CreateTableRequest().withTableName(TABLE_NAME).
                                                                             withKeySchema(new KeySchemaElement().
                                                                                           withAttributeName("name").
                                                                                           withKeyType(KeyType.HASH)).
                                                                             withAttributeDefinitions(new AttributeDefinition().
                                                                                                      withAttributeName("name").
                                                                                                      withAttributeType(ScalarAttributeType.S)).
                                                                             withProvisionedThroughput(new ProvisionedThroughput().
                                                                                                       withReadCapacityUnits(1L).
                                                                                                       withWriteCapacityUnits(1L));

            // Create table if it does not exist yet
            TableUtils.createTableIfNotExists(dynamoDB, createTableRequest);
            //dynamoDB.createTable(createTableRequest);
            // wait for the table to move into ACTIVE state
            TableUtils.waitUntilActive(dynamoDB, TABLE_NAME);

            // Describe our new table
            DescribeTableRequest describeTableRequest = new DescribeTableRequest().withTableName(TABLE_NAME);
            TableDescription tableDescription = dynamoDB.describeTable(describeTableRequest).getTable();
            System.out.println("Tabelleneinstellungen: " + tableDescription);

            // Add an item
            Map<String, AttributeValue> item;
            item = newItem("Van Morrison: The Prophet Speaks", 2018, "*****", "Philipp", "Sarah");
            PutItemRequest putItemRequest = new PutItemRequest(TABLE_NAME, item);
            PutItemResult putItemResult = dynamoDB.putItem(putItemRequest);
            System.out.println("Ergebnis des Einfgens: " + putItemResult);

            // Add another item (intentional error: 2070)
            item = newItem("Joni Mitchell: Shine", 2070, "****", "Rainer", "Benjamin");
            putItemRequest = new PutItemRequest(TABLE_NAME, item);
            putItemResult = dynamoDB.putItem(putItemRequest);
            System.out.println("Ergebnis des Einfgens: " + putItemResult);

            // Update item: correct the error: 2070 => 2007
            item = newItem("Joni Mitchell: Shine", 2007, "****", "Rainer", "Benjamin");
            putItemRequest = new PutItemRequest(TABLE_NAME, item);
            putItemResult = dynamoDB.putItem(putItemRequest);
            System.out.println("Ergebnis des Einfgens: " + putItemResult);

            // Add yet another item
            item = newItem("Joni Mitchell: Clouds", 1969, "****", "Peter", "Paul", "Marie");
            putItemRequest = new PutItemRequest(TABLE_NAME, item);
            putItemResult = dynamoDB.putItem(putItemRequest);
            System.out.println("Ergebnis des Einfgens: " + putItemResult);

            // get item
            HashMap<String, AttributeValue> key = new HashMap<String, AttributeValue>();
            key.put("name", new AttributeValue("Joni Mitchell: Shine"));
            GetItemResult getResult = dynamoDB.getItem(TABLE_NAME, key);
            System.out.println("Ergebnis der Get-Abfrage: " + getResult);

            // query item
            Map<String,String> expressionAttributesNames = new HashMap<>();
            expressionAttributesNames.put("#name","name");
            Map<String,AttributeValue> expressionAttributeValues = new HashMap<>();
            expressionAttributeValues.put(":name", new AttributeValue().withS("Joni Mitchell: Shine"));
            QueryRequest query = new QueryRequest(TABLE_NAME).
                                     withKeyConditionExpression("#name = :name").
                                     withExpressionAttributeNames(expressionAttributesNames).
                                     withExpressionAttributeValues(expressionAttributeValues);
            QueryResult queryResult = dynamoDB.query(query);
            System.out.println("Ergebnis der Query-Abfrage: " + queryResult);

            // Scan items for cds with a year attribute greater than 1985
            HashMap<String, Condition> scanFilter = new HashMap<String, Condition>();
            Condition condition = new Condition().withComparisonOperator(ComparisonOperator.GT.toString()).withAttributeValueList(new AttributeValue().withN("1985"));
            scanFilter.put("year", condition);
            ScanRequest scanRequest = new ScanRequest(TABLE_NAME).withScanFilter(scanFilter);
            ScanResult scanResult = dynamoDB.scan(scanRequest);
            System.out.println("Ergebnis des Scans: " + scanResult);

            try
            {
                Table table = new DynamoDB(dynamoDB).getTable(TABLE_NAME);
                System.out.println("Tabelle wird wieder gelscht ...");
                table.delete();
                table.waitForDelete();
                System.out.print("Erfolg.");

            }
            catch(Exception e)
            {
                System.err.print("Lschen der Tabelle gescheitert: ");
                System.err.println(e.getMessage());
            }
        }

        catch(AmazonServiceException ase)
        {
            System.out.println("AmazonServiceException:");
            System.out.println("   Error Message:    " + ase.getMessage());
            System.out.println("   HTTP Status Code: " + ase.getStatusCode());
            System.out.println("   AWS Error Code:   " + ase.getErrorCode());
            System.out.println("   Error Type:       " + ase.getErrorType());
            System.out.println("   Request ID:       " + ase.getRequestId());
        }
        catch(AmazonClientException ace)
        {
            System.out.println("AmazonClientException:");
            System.out.println("   Error Message: " + ace.getMessage());
        }
    }

    private static Map<String, AttributeValue> newItem(String name, int year, String rating, String... fans)
    {
        Map<String, AttributeValue> item = new HashMap<String, AttributeValue>();
        item.put("name", new AttributeValue(name));
        item.put("year", new AttributeValue().withN(Integer.toString(year)));
        item.put("rating", new AttributeValue(rating));
        item.put("fans", new AttributeValue().withSS(fans));

        return item;
    }
}
