package chapter9.transcribe;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.geometry.Insets;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.layout.*;
import javafx.stage.Stage;
import software.amazon.awssdk.auth.credentials.ProfileCredentialsProvider;
import software.amazon.awssdk.regions.Region;
import software.amazon.awssdk.services.transcribestreaming.TranscribeStreamingAsyncClient;
import software.amazon.awssdk.services.transcribestreaming.model.*;

public class Gui extends Application
{
    private static final String BUTTON_LABEL_START = "Whrend des Sprechens gedrckt halten";
    private static final String BUTTON_LABEL_RUNNING = "Aufnahme luft";
    private static final int SAMPLE_RATE = 16_000;

    private Button recordingButton;
    private InputStream input;
    
    public void start(Stage primaryStage)
    {
        VBox root = new VBox(10);
        root.setPadding(new Insets(10));

        TextArea textArea = new TextArea();
        textArea.setMaxHeight(Double.MAX_VALUE);
        textArea.setWrapText(true);
        VBox.setVgrow(textArea, Priority.ALWAYS);
        
        HBox commandLine = new HBox(10);
        ComboBox<Language> language = new ComboBox<>();
        language.getItems().addAll(Language.values());
        language.getSelectionModel().select(0);
        recordingButton = new Button(BUTTON_LABEL_START);
        Button clearButton = new Button("Inhalt lschen");
        commandLine.getChildren().addAll(language, recordingButton, clearButton);
        
        root.getChildren().addAll(commandLine, textArea);

        recordingButton.setOnMousePressed(e -> 
                                     startRecording(textArea, 
                                                    language.getSelectionModel().getSelectedItem()));
        recordingButton.setOnMouseReleased(e -> stopRecording());
        clearButton.setOnAction(e -> textArea.clear());
        
        Scene scene = new Scene(root, 1000, 250);
        primaryStage.setScene(scene);
        primaryStage.setTitle("Sprache zu Text");
        
        primaryStage.setOnCloseRequest(e -> {System.exit(0);});
        primaryStage.show();
    }

    private void startRecording(TextArea area, Language language)
    {
        if(language == null)
        {
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle("Fehler");
            alert.setHeaderText("Sprache nicht gewhlt");
            alert.setHeaderText(null);
            alert.setContentText("Es wurde keine Sprache ausgewhlt!");
            alert.showAndWait();
            return;
        }
        recordingButton.setPrefWidth(recordingButton.getWidth());
        recordingButton.setText(BUTTON_LABEL_RUNNING);
        TranscribeStreamingAsyncClient client = TranscribeStreamingAsyncClient.builder()
                                                .credentialsProvider(ProfileCredentialsProvider.create("Transcribe"))
                                                .region(Region.EU_CENTRAL_1).build();

        StartStreamTranscriptionRequest request = StartStreamTranscriptionRequest.builder()
                                                  .languageCode(language.getLanguageCode().toString())
                                                  .mediaEncoding(MediaEncoding.PCM)
                                                  .mediaSampleRateHertz(SAMPLE_RATE).build();
        try
        {
            input = getStreamFromMic();
            if(input == null)
            {
                throw new LineUnavailableException();
            }
        }
        catch(LineUnavailableException e)
        {
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle("Fehler");
            alert.setHeaderText("Problem mit Mikrofon");
            alert.setHeaderText(null);
            alert.setContentText("Mikrofoneingabe kann nicht geffnet werden!");
            alert.showAndWait();
            return;
        }
        AudioStreamPublisher publisher = new AudioStreamPublisher(input);
        client.startStreamTranscription(request, publisher, getResponseHandler(area));
    }

    private static InputStream getStreamFromMic() throws LineUnavailableException
    {
        AudioFormat format = new AudioFormat(SAMPLE_RATE, 16, 1, true, false);
        DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);

        if(!AudioSystem.isLineSupported(info))
        {
            return null;
        }

        TargetDataLine line = (TargetDataLine) AudioSystem.getLine(info);
        line.open(format);
        line.start();

        InputStream audioStream = new AudioInputStream(line);
        return audioStream;
    }

    private static StartStreamTranscriptionResponseHandler getResponseHandler(TextArea area)
    {
        return StartStreamTranscriptionResponseHandler.builder()
                        .onResponse(r -> {Platform.runLater(()->area.appendText("=== Aufnahme beginnt ===\n"));})
                        .onError(e -> {Platform.runLater(()->area.appendText("=== FEHLER (s. Console) ===\n"));
                                       e.printStackTrace();})
                        .onComplete(() -> {Platform.runLater(()->area.appendText("=== Aufnahme beendet ===\n"));})
                        .subscriber(event -> 
                            {
                                List<Result> results = ((TranscriptEvent) event).transcript().results();
                                if(results.size() > 0)
                                {
                                    Platform.runLater(()->append(area, results.get(0).isPartial(),
                                                                 results.get(0).alternatives().get(0).transcript()));
                                }
                            })
                        .build();
    }
    
    private static void append(TextArea area, boolean partial, String newText)
    {
        //partial text: ohne Newline, der jedes Mal wieder entfernt wird
        //non-partial text: mit Newline, bleibt stehen
        String text = area.getText();
        if(text.length() > 0 && text.charAt(text.length()-1) != '\n')
        {
            int lastNewLinePos = text.lastIndexOf('\n');
            if(lastNewLinePos >= 0)
            {
                area.deleteText(lastNewLinePos+1, text.length());
            }
            //else: berhaupt kein Newline vorhanden => nichts da zu lschen
        }
        //else: kein Text da oder letzte Zeile endet mit Newline => nichts da zu lschen
        String lineEnd = partial? "" : "\n";
        area.appendText(newText + lineEnd);
    }

    private void stopRecording()
    {
        recordingButton.setText(BUTTON_LABEL_START);
        if(input != null)
        {
            try
            {
                input.close();
            }
            catch(IOException e)
            {
            }
        }
    }

    public static void main(String[] args)
    {
        launch(args);
    }
}
